#include "stdafx.h"
#include "cofpcdat.h"
#include "exceptions.h"
#include "hexdump.h"

/*
 * BEGIN IMAGE_SECTION_HELPER
 */
byte * image_section_helper::get_rva(unsigned long addr)
{
	if (addr < start || addr >= end) throw std::runtime_error("Invalid RVA for section");
	return mbase + (addr - start);
}
int image_section_helper::section_id() 
{
	return mid;
}
unsigned long image_section_helper::dumpsize()
{
	unsigned long size = mh->Misc.VirtualSize ? mh->Misc.VirtualSize : mh->SizeOfRawData;
	if (mh->Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA || mh->Characteristics & IMAGE_SCN_CNT_CODE)
//			&& (mh->Characteristics & (IMAGE_SCN_MEM_SHARED|IMAGE_SCN_MEM_EXECUTE|IMAGE_SCN_MEM_WRITE)))
	{
		if (mh->Misc.VirtualSize == 0 && mh->SizeOfRawData)
		{
			size = mh->SizeOfRawData;
		}
		else
		{
			size = min(mh->Misc.VirtualSize,mh->SizeOfRawData);
		}
	}
	return size;
}
unsigned long image_section_helper::summarysize(unsigned long alignment)
{
	unsigned long size = mh->Misc.VirtualSize ? mh->Misc.VirtualSize : mh->SizeOfRawData;
	if (mh->Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA
		&& (0 == (mh->Characteristics & (IMAGE_SCN_MEM_EXECUTE|IMAGE_SCN_MEM_WRITE)))) size = mh->SizeOfRawData;
	if (!alignment) return size;
	if ((size & (alignment-1)) == 0) return size;
	size += alignment;
	size &= ~(alignment -1);
	return size;
}
std::string image_section_helper::name() const
{
	return str(mh->Name,IMAGE_SIZEOF_SHORT_NAME);
}
unsigned long image_section_helper::characteristics() const
{
	return mh->Characteristics;
}
bool image_section_helper::uses_virtual_size()
{
	return false;
}
/* 
 * END IMAGE_SECTION_HELPER
 */ 


/* 
 * BEGIN COFPCDAT
 */ 
/*
 * Preprocess the symbol table to find overrides/aliases
 */
void cofpcdat::preprocess_symtable()
{
	if (psym)
	{
		for (int i = 0; i < nsym; ++i)
		{
			IMAGE_SYMBOL & s = psym[i];
			if (s.Type == 32)
			{
				maliases[s.Value] = symbol_name(s, this->stringtable);
			}
			i += psym[i].NumberOfAuxSymbols;
		}
	}
}
bool cofpcdat::test_symtable_alias(unsigned long idx, std::string & result)
{
	std::map<unsigned long, std::string>::iterator mit = maliases.find(idx);
	if (mit == maliases.end()) return false;
	result = mit->second;
	return true;
}
/*
 * END COFPCDAT.
 */


/*
 * Dumpbin orders the data directories oddly.  Let's replicate!
 */
std::map<int, int> cofpcdat::ddorder()
{
	std::map<int, int> res;
	int idx = 0;
	res[IMAGE_DIRECTORY_ENTRY_DEBUG]			= idx++;
	res[IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR]	= idx++;
	res[IMAGE_DIRECTORY_ENTRY_EXPORT]			= idx++;
	res[IMAGE_DIRECTORY_ENTRY_IMPORT]			= idx++;
	res[IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_RESOURCE]			= idx++;
	res[IMAGE_DIRECTORY_ENTRY_EXCEPTION]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_SECURITY]			= idx++;
	res[IMAGE_DIRECTORY_ENTRY_BASERELOC]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_ARCHITECTURE]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_GLOBALPTR]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG]		= idx++;
	res[IMAGE_DIRECTORY_ENTRY_TLS]				= idx++;
	res[IMAGE_DIRECTORY_ENTRY_IAT]				= idx++;

	return res;
}

std::string cofpcdat::stringtable_string(std::string & ref) 
{
	if (ref[0] != '/') return "";
	return std::string((char*)stringtable + atoi(ref.c_str()+1));
}
// add open interval.
void cofpcdat::add_section_information(IMAGE_SECTION_HEADER * im, int sectionid)
{
	image_section_helper sh(im, sectionid, this->base);
	
	m_sectbyname[sh.name()].push_back(sh);

	unsigned long start = im->VirtualAddress;
	unsigned long end   = im->VirtualAddress + im->Misc.VirtualSize - 1;
	rvarangebound rb;
	rb.sectionid = sectionid;
	rb.start = true;
	rb.sectionheader = im;
	m_rvaranges[start] = rb;
	rb.start = false;
	m_rvaranges[end] = rb;
}
bool cofpcdat::test_rva(unsigned long addr)
{
	rvaranges::iterator rit = m_rvaranges.lower_bound(addr);
	if (rit == m_rvaranges.end()) return false;
	rvarangebound & rv = rit->second;
	if (addr < rv.sectionheader->VirtualAddress) return false;
	return true;
}

int cofpcdat::find_rva_section(unsigned long addr)
{
	rvaranges::iterator rit = m_rvaranges.lower_bound(addr);
	if (rit == m_rvaranges.end()) throw invalid_rva_error("Invalid RVA!");
	rvarangebound & rv = rit->second;
	if (addr < rv.sectionheader->VirtualAddress) throw invalid_rva_error("Invalid RVA! (too small!)");
	return rv.sectionid;
}
image_section_helper cofpcdat::get_rva_section(unsigned long addr)
{
	rvaranges::iterator rit = m_rvaranges.lower_bound(addr);
	if (rit == m_rvaranges.end()) throw std::runtime_error("Invalid RVA!");
	rvarangebound & rv = rit->second;
	if (addr < rv.sectionheader->VirtualAddress) throw invalid_rva_error("Invalid RVA! (too small!)");
	return image_section_helper(rv.sectionheader, rv.sectionid, this->base);
}
byte * cofpcdat::get_rva(unsigned long addr)
{
	rvaranges::iterator rit = m_rvaranges.lower_bound(addr);
	if (rit == m_rvaranges.end()) throw std::runtime_error("Invalid RVA!");
	
	rvarangebound & rv = rit->second;
	if (addr < rv.sectionheader->VirtualAddress) throw invalid_rva_error("Invalid RVA! (too small!)");
	byte * imbase = this->base + rv.sectionheader->PointerToRawData;
	unsigned long offset = addr - rv.sectionheader->VirtualAddress;
	if (offset >= rv.sectionheader->SizeOfRawData)
	{
		throw invalid_rva_error("Valid rva, but not in raw data.");
	}
	imbase += offset;
	return imbase;
}
void cofpcdat::add_data_directory(int ddid, IMAGE_DATA_DIRECTORY * pd)
{
	int sectionid;
	try {
		sectionid = find_rva_section(pd->VirtualAddress);
	}
	catch (invalid_rva_error &)
	{
		if (ddid == IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT) 
		{
			// which is the .rdata section?
			sectionsnamemap::iterator mit = m_sectbyname.find(".rdata");
			if (mit == m_sectbyname.end()) mit = m_sectbyname.find(".text");
			if (mit == m_sectbyname.end()) 
			{
				sectionid = 1;
			}
			else
			{
				sectionid = mit->second.front().section_id();
			}
		}
		else
		{
			throw;
		}
	}

	datadirectory d;
	d.ddid = ddid;
	d.pdd  = pd;
	
	m_ddowners[sectionid].insert(d);
}
bool cofpcdat::has_data_directories(int sectionid)
{
	return (m_ddowners.find(sectionid) != m_ddowners.end());
}

cofpcdat::datadirectoryvec & cofpcdat::get_data_directories(int sectionid)
{
	datadirectoryvec ddv = m_ddowners[sectionid];
	// debugview has an odd ordering!
	
	return m_ddowners[sectionid];
}
cofpcdat::sectionsnamemap & cofpcdat::get_sectionsbyname()
{
	return m_sectbyname;
}

std::map<int, int> cofpcdat::dd_cmp::dorder = cofpcdat::ddorder();
/* 
 * END COFPCDAT
 */ 
