#include "stdafx.h"
#include "hexdump.h"
#include <sstream>
#include "unmangle.h"
#include "dump_helpers.h"

std::string machine_type(short m)
{
	switch (m)
	{
	case IMAGE_FILE_MACHINE_I386:
		return "x86";
	case IMAGE_FILE_MACHINE_ARM:
		return "ARM";
	case IMAGE_FILE_MACHINE_R3000:
	case IMAGE_FILE_MACHINE_R4000:
	case IMAGE_FILE_MACHINE_R10000:
	case IMAGE_FILE_MACHINE_WCEMIPSV2:
	case IMAGE_FILE_MACHINE_ALPHA:    
	case IMAGE_FILE_MACHINE_SH3:      
	case IMAGE_FILE_MACHINE_SH3DSP:   
	case IMAGE_FILE_MACHINE_SH3E:     
	case IMAGE_FILE_MACHINE_SH4:      
	case IMAGE_FILE_MACHINE_SH5:      
	case IMAGE_FILE_MACHINE_THUMB:    
	case IMAGE_FILE_MACHINE_AM33:     
	case IMAGE_FILE_MACHINE_POWERPC:  
	case IMAGE_FILE_MACHINE_POWERPCFP:
	case IMAGE_FILE_MACHINE_IA64:     
	case IMAGE_FILE_MACHINE_MIPS16:   
	case IMAGE_FILE_MACHINE_ALPHA64:  
	case IMAGE_FILE_MACHINE_MIPSFPU:  
	case IMAGE_FILE_MACHINE_MIPSFPU16:
	case IMAGE_FILE_MACHINE_TRICORE:  
	case IMAGE_FILE_MACHINE_CEF:      
	case IMAGE_FILE_MACHINE_EBC:      
	case IMAGE_FILE_MACHINE_AMD64:    
	case IMAGE_FILE_MACHINE_M32R:     
	case IMAGE_FILE_MACHINE_CEE:
		return "NYI";
	default:
		return "???";
	}
}

std::string dump_ddinfo(IMAGE_DATA_DIRECTORY * idd, const char * name)
{
	unsigned long va   = idd->VirtualAddress;
	unsigned long size = idd->Size;
	std::stringstream ss;
	ss << lpad(hex(va), 16) << " [" << lpad(hex(size), 8) << "] RVA [size] of " << name << " Directory\n";
	return ss.str();
}

std::string import_object_header_type(int t)
{
	// bounds check.
	char * types[] = { "code", "data", "const" };
	return types[t];
}

std::string import_object_header_name_type(int t)
{
    //IMPORT_OBJECT_ORDINAL = 0,          // Import by ordinal
    //IMPORT_OBJECT_NAME = 1,             // Import name == public symbol name.
    //IMPORT_OBJECT_NAME_NO_PREFIX = 2,   // Import name == public symbol name skipping leading ?, @, or optionally _.
    //IMPORT_OBJECT_NAME_UNDECORATE = 3,  // Import name == public symbol name skipping leading ?, @, or optionally _
    //                                    // and truncating at first @

	// bounds check.
	char * types[] = { "ordinal", "name", "no prefix", "undecorate" };
	return types[t];
}

std::string data_directory_name(int ddid)
{
	char* names[] = {"Export", "Import", "Resource", "Exception", "Certificates", "Base Relocation", "Debug", "Architecture",
					 "Global Pointer", "Thread Storage", "Load Configuration", "Bound Import", "Import Address Table",
					 "Delay Import", "COM Descriptor", "Reserved", 0};
	return std::string(names[ddid]);
}

std::string symbol_storageclass(int i, std::string & s)
{
	switch (i)
	{
	case IMAGE_SYM_CLASS_FUNCTION:			
		{
			if (s == ".ef") return "EndFunction";
			if (s == ".bf") return "BeginFunction";
			return ".bf or.ef";
		}
	case IMAGE_SYM_CLASS_END_OF_FUNCTION:	return "EndFunction"; 
	case IMAGE_SYM_CLASS_NULL:				return "Null"; 
	case IMAGE_SYM_CLASS_AUTOMATIC:			return "Automatic"; 
	case IMAGE_SYM_CLASS_EXTERNAL:			return "External"; 
	case IMAGE_SYM_CLASS_STATIC:			return "Static"; 
	case IMAGE_SYM_CLASS_LABEL:				return "Label";
	case IMAGE_SYM_CLASS_SECTION:			return "Section";
	case IMAGE_SYM_CLASS_WEAK_EXTERNAL:		return "WeakExternal";
	case IMAGE_SYM_CLASS_FILE:				return "Filename";
	default: return str(i);
	}
}

std::string symbol_type(int i)
{
	switch (i)
	{
	case IMAGE_SYM_TYPE_NULL: return  "notype"; 
	case 32:				  return  "notype ()";
	case IMAGE_SYM_TYPE_VOID: return  "VOID"; 
	case IMAGE_SYM_TYPE_CHAR: return  "CHAR"; 
	case IMAGE_SYM_TYPE_SHORT: return  "SHORT"; 
	case IMAGE_SYM_TYPE_INT: return  "INT"; 
	case IMAGE_SYM_TYPE_LONG: return  "LONG"; 
	case IMAGE_SYM_TYPE_FLOAT: return  "FLOAT"; 
	case IMAGE_SYM_TYPE_DOUBLE: return  "DOUBLE"; 
	case IMAGE_SYM_TYPE_STRUCT: return  "STRUCT"; 
	case IMAGE_SYM_TYPE_UNION: return  "UNION"; 
	case IMAGE_SYM_TYPE_ENUM: return  "ENUM"; 
	case IMAGE_SYM_TYPE_MOE: return  "MOE"; 
	case IMAGE_SYM_TYPE_BYTE: return  "BYTE"; 
	case IMAGE_SYM_TYPE_WORD: return  "WORD"; 
	case IMAGE_SYM_TYPE_UINT: return  "UINT"; 
	case IMAGE_SYM_TYPE_DWORD: return  "DWORD"; 
	case IMAGE_SYM_TYPE_PCODE: return  "PCODE"; 
	default: return  str(i); 
	}
}

std::string print_symbol(IMAGE_SYMBOL & s, byte * stringtable, int i)
{
	std::stringstream ssc;

	std::stringstream ss;

	std::string sidx = lpad(hex(i), 3, '0');
	ss << sidx << " ";
	ss << lpad(hex(str(s.Value)), 8, '0') << " ";

	switch (s.SectionNumber) 
	{
	case -2:
		ss << "DEBUG  "; break;
	case -1:
		ss << "ABS    "; break;
	case 0:
		ss << "UNDEF  "; break;
	default:
		ss << "SECT" << rpad(hex(s.SectionNumber),2) << " "; break;
	}
	ss << rpad(symbol_type(s.Type),13);
	std::string name = symbol_name(s, stringtable);

	ss << rpad(symbol_storageclass(s.StorageClass, name), 12);

	ssc << rpad(ss.str(), 46 + ((int)sidx.length() - 3));

	// The previous is padded as a whole!! :(
	ssc << "| ";

	ssc << name;
	std::string umname = unmangle(name);
	if (umname != "") ssc << " (" << unmangle(name) << ")";
	ssc << "\n";

	return ssc.str();
}


std::string dump_aux_file(IMAGE_SYMBOL & sym, IMAGE_AUX_SYMBOL * paux, int naux, byte * stringtable)
{
	if (naux != 1) return "";
	std::stringstream ssc;
	ssc << "    " << str(paux->File.Name, IMAGE_SIZEOF_SYMBOL) << "\n";
	return ssc.str();
}

std::string dump_aux_weak_externals(IMAGE_SYMBOL & sym, IMAGE_AUX_SYMBOL * paux, int naux)
{
	if (naux != 1) return "";
	std::stringstream ssc;
	ssc << "    Default index " << lpad(hex(paux->Sym.TagIndex), 8) << " library search\n";
	return ssc.str();
}

std::string dump_aux_externals(IMAGE_SYMBOL & sym, IMAGE_AUX_SYMBOL * paux, int naux)
{
	if (naux != 1) return "";
	std::stringstream ssc;
	ssc       << "    tag index " << lpad(hex(paux->Sym.TagIndex), 8, '0') <<
			     " size " << lpad(hex(paux->Sym.Misc.TotalSize), 8, '0') <<
				 " lines " << lpad(hex(paux->Sym.FcnAry.Function.PointerToLinenumber), 8, '0') <<
				 " next function " << lpad(hex(paux->Sym.FcnAry.Function.PointerToNextFunction), 8, '0') << "\n";
	return ssc.str();
}

std::string dump_aux_function(IMAGE_SYMBOL & sym, IMAGE_AUX_SYMBOL * paux, int naux, std::string & name)
{
	if (naux != 1) return "";
	std::stringstream ssc;
	ssc << "    line# " << lpad(hex(paux->Sym.Misc.LnSz.Linenumber), 4, '0');
	if (name == ".bf") ssc << " end " << lpad(hex(paux->Sym.Misc.LnSz.Size), 8, '0');
	ssc << "\n";
	return ssc.str();
}


std::string dump_aux_statics(IMAGE_SYMBOL & sym, IMAGE_AUX_SYMBOL * paux, int naux)
{
	std::stringstream ssc;
	if (naux < 1) return "";
	ssc << "    Section length " << lpad(hex(paux->Section.Length), 4);
	ssc << ", #relocs " << lpad(hex(paux->Section.NumberOfRelocations), 4);
	ssc << ", #linenums " << lpad(hex(paux->Section.NumberOfLinenumbers), 4);
	ssc << ", checksum " << lpad(hex((unsigned int)paux->Section.CheckSum), 8);

	ssc << dump_comdat_selection(paux->Section.Selection, paux->Section.Number);
	if (naux < 2) return ssc.str();

	ssc << "    Relocation CRC " << lpad(hex(paux[1].Section.Length),8,'0') << "\n";
	return ssc.str();
}

std::string dump_comdat_selection(int selection, int section)
{
	std::stringstream ssc;

	if (!selection) 
	{
		return "\n";
	}

	ssc << ", selection ";
	switch (selection)
	{
	case IMAGE_COMDAT_SELECT_NODUPLICATES:
		ssc << "   1 (pick no duplicates"; break;
	case IMAGE_COMDAT_SELECT_ANY:
		ssc << "   2 (pick any"; break;
	case IMAGE_COMDAT_SELECT_SAME_SIZE:
		ssc << "   3( pick same size"; break;
	case IMAGE_COMDAT_SELECT_EXACT_MATCH:
		ssc << "   4 (pick exact match"; break;
	case IMAGE_COMDAT_SELECT_ASSOCIATIVE:
		ssc << "   5 (pick associative Section 0x" << hex(section); break;
	case IMAGE_COMDAT_SELECT_LARGEST:
		ssc << "   6 (pick largest"; break;
	case IMAGE_COMDAT_SELECT_NEWEST:
		ssc << "   7 (pick newest"; break;
	}
	ssc << ")\n";

	return ssc.str();
}